import argparse
import datetime
import os
import sys
import time
import traceback
from importlib import metadata
from importlib.metadata import PackageMetadata

import yaml

from slack_watchman import (
    sw_logger,
    signature_downloader,
    exceptions,
    slack_wrapper
)
from slack_watchman.models import (
    signature,
    user,
    workspace,
    post,
    conversation
)

OUTPUT_LOGGER: sw_logger.JSONLogger


def validate_conf(path: str, cookie: bool) -> bool:
    """ Check the file slack_watchman.conf exists

    Args:
        path: Path for the .config file
        cookie: Whether session:cookie auth is being used
    Returns:
        True if the required config settings are present, False if not
    """

    if not cookie:
        if not os.environ.get('SLACK_WATCHMAN_TOKEN'):
            if os.path.exists(f'{os.path.expanduser("~")}/slack_watchman.conf'):
                OUTPUT_LOGGER.log('WARNING', 'Legacy slack_watchman.conf file detected. Renaming to watchman.conf')
                os.rename(rf'{os.path.expanduser("~")}/slack_watchman.conf',
                          rf'{os.path.expanduser("~")}/watchman.conf')
                try:
                    with open(path) as yaml_file:
                        return yaml.safe_load(yaml_file)['slack_watchman']
                except:
                    raise exceptions.MisconfiguredConfFileError
            elif os.path.exists(path):
                try:
                    with open(path) as yaml_file:
                        return yaml.safe_load(yaml_file)['slack_watchman']
                except:
                    raise exceptions.MisconfiguredConfFileError
            else:
                try:
                    os.environ['SLACK_WATCHMAN_TOKEN']
                except:
                    raise exceptions.MissingEnvVarError('SLACK_WATCHMAN_TOKEN')
    else:
        if os.path.exists(f'{os.path.expanduser("~")}/slack_watchman.conf'):
            OUTPUT_LOGGER.log('WARNING', 'Legacy slack_watchman.conf file detected. Renaming to watchman.conf')
            os.rename(rf'{os.path.expanduser("~")}/slack_watchman.conf',
                      rf'{os.path.expanduser("~")}/watchman.conf')
            try:
                with open(path) as yaml_file:
                    return yaml.safe_load(yaml_file)['slack_watchman']
            except:
                raise exceptions.MisconfiguredConfFileError
        elif os.path.exists(path):
            try:
                with open(path) as yaml_file:
                    return yaml.safe_load(yaml_file)['slack_watchman']
            except:
                raise exceptions.MisconfiguredConfFileError
        else:
            try:
                os.environ['SLACK_WATCHMAN_COOKIE']
            except:
                raise exceptions.MissingEnvVarError('SLACK_WATCHMAN_COOKIE')

            try:
                os.environ['SLACK_WATCHMAN_URL']
            except:
                raise exceptions.MissingEnvVarError('SLACK_WATCHMAN_URL')


def search(slack_connection: slack_wrapper.SlackAPI,
           loaded_signature: signature.Signature,
           timeframe: int or str,
           scope: str,
           verbose: bool) -> None:
    """ Search either messages or files for matches from signatures.
    Findings are then filtered by RegEx, and any matches are output to
    the chosen logging mechanism.

    Args:
        slack_connection: Slack API object
        loaded_signature: Signature object which defines what to search for
        timeframe: How far back to search for
        scope: Where to search in Slack, e.g. files or messages
        verbose: Whether to use verbose logging or not
    """

    if scope == 'messages':
        OUTPUT_LOGGER.log('INFO', f'Searching for posts containing {loaded_signature.name}')
        messages = slack_wrapper.find_messages(
            slack_connection,
            OUTPUT_LOGGER,
            loaded_signature,
            verbose,
            timeframe)
        if messages:
            for message in messages:
                OUTPUT_LOGGER.log(
                    'NOTIFY',
                    message,
                    scope=scope,
                    severity=loaded_signature.severity,
                    detect_type=loaded_signature.name,
                    notify_type='result')
    if scope == 'files':
        OUTPUT_LOGGER.log('INFO', f'Searching for posts containing {loaded_signature.name}')
        files = slack_wrapper.find_files(
            slack_connection,
            OUTPUT_LOGGER,
            loaded_signature,
            verbose,
            timeframe)
        if files:
            for file in files:
                OUTPUT_LOGGER.log(
                    'NOTIFY',
                    file,
                    scope=scope,
                    severity=loaded_signature.severity,
                    detect_type=loaded_signature.name,
                    notify_type='result')


def unauthenticated_probe(workspace_domain: str,
                          project_metadata: PackageMetadata) -> None:
    """ Probe Slack for information about the workspace without authentication.

    Args:
        workspace_domain: Domain of the workspace
        project_metadata: Project metadata
    """

    OUTPUT_LOGGER.log('SUCCESS', 'Slack Watchman started execution')
    OUTPUT_LOGGER.log('INFO', f'Version: {project_metadata.get("version")}')
    OUTPUT_LOGGER.log('INFO', f'Created by: PaperMtn <papermtn@protonmail.com>')
    OUTPUT_LOGGER.log('SUCCESS', f'Running in probe mode')
    OUTPUT_LOGGER.log('SUCCESS', f'Slack Watchman will attempt an unauthenticated probe on the workspace '
                                 f'and return any available authentication information.')
    OUTPUT_LOGGER.log('SUCCESS', f'Workspace: {workspace_domain}')
    try:
        domain_information = slack_wrapper.find_auth_information(workspace_domain)
        if domain_information:
            OUTPUT_LOGGER.log('WORKSPACE_PROBE', domain_information, detect_type='Workspace Probe',
                              notify_type='workspace_probe')
            OUTPUT_LOGGER.log('SUCCESS', 'Slack Watchman probe finished execution. '
                                         'Run in full mode with token authentication to scan a workspace')
        else:
            OUTPUT_LOGGER.log('INFO', f'No information found for the workspace {workspace_domain}. '
                                      f'This may not be a Slack Workspace, or there may be no authentication'
                                      f' information available.')
            OUTPUT_LOGGER.log('SUCCESS', 'Slack Watchman probe finished execution. '
                                         'Run in full mode with token authentication to scan a workspace')
        sys.exit()
    except SystemExit:
        sys.exit(1)
    except Exception as e:
        OUTPUT_LOGGER.log('CRITICAL', e)
        sys.exit(1)


def init_logger(logging_type: str, debug: bool) -> sw_logger.JSONLogger or sw_logger.StdoutLogger:
    """ Create a logger object. Defaults to stdout if no option is given

    Args:
        logging_type: Type of logging to use
        debug: Whether to use debug level logging or not
    Returns:
        Logger object
    """

    if not logging_type or logging_type == 'stdout':
        return sw_logger.StdoutLogger(debug=debug)
    else:
        return sw_logger.JSONLogger(debug=debug)


def main():
    global OUTPUT_LOGGER
    try:
        OUTPUT_LOGGER = ''
        start_time = time.time()
        project_metadata = metadata.metadata('slack-watchman')
        parser = argparse.ArgumentParser(description="Monitoring and enumerating Slack for exposed secrets")

        parser.add_argument('--timeframe', '-t', choices=['d', 'w', 'm', 'a'], dest='time', default='a',
                            help='How far back to search: d = 24 hours w = 7 days, m = 30 days, a = all time')
        parser.add_argument('--output', '-o', choices=['json', 'stdout'], dest='logging_type',
                            help='Where to send results')
        parser.add_argument('--version', '-v', action='version',
                            version=f'Slack Watchman: {project_metadata.get("version")}')
        parser.add_argument('--all', '-a', dest='everything', action='store_true',
                            help='Find secrets and PII')
        parser.add_argument('--users', '-u', dest='users', action='store_true',
                            help='Enumerate users and output them to .csv in the current working directory')
        parser.add_argument('--channels', '-c', dest='channels', action='store_true',
                            help='Enumerate channels and output them to .csv in the current working directory')
        parser.add_argument('--pii', '-p', dest='pii', action='store_true',
                            help='Find personal data: DOB, passport details, drivers licence, ITIN, SSN etc.')
        parser.add_argument('--secrets', '-s', dest='secrets', action='store_true',
                            help='Find exposed secrets: credentials, tokens etc.')
        parser.add_argument('--financial', dest='financial', action='store_true', help=argparse.SUPPRESS)
        parser.add_argument('--tokens', dest='tokens', action='store_true', help=argparse.SUPPRESS)
        parser.add_argument('--files', dest='files', action='store_true', help=argparse.SUPPRESS)
        parser.add_argument('--custom', dest='custom', action='store_true', help=argparse.SUPPRESS)
        parser.add_argument('--debug', '-d', dest='debug', action='store_true', help='Turn on debug level logging')
        parser.add_argument('--verbose', '-V', dest='verbose', action='store_true',
                            help='Turn on more verbose output for JSON logging. '
                                 'This includes more fields, but is larger')
        parser.add_argument('--cookie', dest='cookie', action='store_true',
                            help='Use cookie auth using Slack d cookie. '
                                 'REQUIRES either SLACK_WATCHMAN_COOKIE and SLACK_WATCHMAN_URL environment variables '
                                 'set, or both values set in watchman.conf')
        parser.add_argument('--probe', dest='probe_domain',
                            help='Perform an un-authenticated probe on a workspace for available'
                                 ' authentication options and other information. '
                                 'Enter workspace domain to probe')

        args = parser.parse_args()
        tm = args.time
        everything = args.everything
        users = args.users
        channels = args.channels
        logging_type = args.logging_type
        debug = args.debug
        verbose = args.verbose
        secrets = args.secrets
        pii = args.pii
        cookie = args.cookie
        probe_domain = args.probe_domain

        OUTPUT_LOGGER = init_logger(logging_type, debug)

        for deprecated_arg in ['financial', 'tokens', 'files', 'custom']:
            if vars(args).get(deprecated_arg):
                OUTPUT_LOGGER.log('WARNING', f'Argument `--{deprecated_arg}` is deprecated, and has been ignored')

        if tm == 'd':
            now = int(time.time())
            timeframe = time.strftime('%Y-%m-%d', time.localtime(now - 86400))
        elif tm == 'w':
            now = int(time.time())
            timeframe = time.strftime('%Y-%m-%d', time.localtime(now - 604800))
        elif tm == 'm':
            now = int(time.time())
            timeframe = time.strftime('%Y-%m-%d', time.localtime(now - 2592000))
        else:
            now = int(time.time())
            timeframe = time.strftime('%Y-%m-%d', time.localtime(now - 1576800000))

        if probe_domain:
            unauthenticated_probe(probe_domain, project_metadata)

        conf_path = f'{os.path.expanduser("~")}/watchman.conf'
        validate_conf(conf_path, cookie)
        slack_con = slack_wrapper.initiate_slack_connection(cookie)

        auth_data = slack_con.get_auth_test()
        calling_user = user.create_from_dict(
            slack_con.get_user_info(auth_data.get('user_id')).get('user'), True)

        workspace_information = workspace.create_from_dict(slack_con.get_workspace_info().get('team'))

        OUTPUT_LOGGER.log('SUCCESS', 'Slack Watchman started execution')
        OUTPUT_LOGGER.log('INFO', f'Version: {project_metadata.get("version")}')
        OUTPUT_LOGGER.log('INFO', f'Created by: PaperMtn <papermtn@protonmail.com>')
        OUTPUT_LOGGER.log('INFO', f'Searching workspace: {workspace_information.name}')
        OUTPUT_LOGGER.log('INFO', f'Workspace URL: {workspace_information.url}')
        OUTPUT_LOGGER.log('INFO', 'Downloading and importing signatures...')
        signature_list = signature_downloader.SignatureDownloader(OUTPUT_LOGGER).download_signatures()
        OUTPUT_LOGGER.log('SUCCESS', f'{len(signature_list)} signatures loaded')
        if cookie:
            OUTPUT_LOGGER.log('SUCCESS', 'Successfully authenticated using cookie')
            OUTPUT_LOGGER.log('SUCCESS', f"This user's SESSION_TOKEN: {slack_con.session_token}")
        OUTPUT_LOGGER.log('SUCCESS',
                          f'You are authenticated to this workspace as: USER: {calling_user.display_name} '
                          f'- {calling_user.email} ID: {calling_user.id}')
        OUTPUT_LOGGER.log('USER', calling_user, detect_type='User', notify_type='user')
        OUTPUT_LOGGER.log('WORKSPACE', workspace_information, detect_type='Workspace', notify_type='workspace')
        OUTPUT_LOGGER.log('INFO', 'Finding workspace authentication options')
        workspace_auth = slack_wrapper.find_auth_information(domain_url=workspace_information.url)
        if workspace_auth:
            OUTPUT_LOGGER.log('WORKSPACE_AUTH', workspace_auth, detect_type='Workspace Auth',
                              notify_type='workspace_auth')
        else:
            OUTPUT_LOGGER.log('INFO', 'No workspace authentication information found')

        if users:
            OUTPUT_LOGGER.log('INFO', 'Enumerating users...')
            user_list = slack_wrapper.get_users(slack_con, verbose)
            OUTPUT_LOGGER.log('SUCCESS', f'{len(user_list)} users discovered')
            OUTPUT_LOGGER.log('INFO', 'Writing to csv')
            sw_logger.export_csv('slack_users', user_list)
            OUTPUT_LOGGER.log(
                'SUCCESS',
                f'Users output to CSV file: {os.path.join(os.getcwd(), "slack_users.csv")}')

        if channels:
            OUTPUT_LOGGER.log('INFO', 'Enumerating channels...')
            channel_list = slack_wrapper.get_channels(slack_con, verbose)
            OUTPUT_LOGGER.log('SUCCESS', f'{len(channel_list)} channels discovered')
            OUTPUT_LOGGER.log('INFO', 'Writing to csv')
            sw_logger.export_csv('slack_channels', channel_list)
            OUTPUT_LOGGER.log(
                'SUCCESS',
                f'Users output to CSV file: {os.path.join(os.getcwd(), "slack_channels.csv")}')
            OUTPUT_LOGGER.log('INFO', 'Finding public Canvases')
            for channel in channel_list:
                if not channel.canvas_empty and channel.canvas_id:
                    canvas_information = {
                        'channel_name': channel.name,
                        'canvas_url': f'{workspace_information.url}canvas/{channel.id}'
                    }
                    OUTPUT_LOGGER.log('CANVAS', canvas_information, detect_type='Canvas',
                                      notify_type='canvas')
        if everything or not pii and not secrets:
            OUTPUT_LOGGER.log('INFO', 'Searching for PII and Secrets')
            for signature in signature_list:
                for scope in signature.scope:
                    search(
                        slack_con,
                        signature,
                        timeframe,
                        scope,
                        verbose)
        elif secrets:
            OUTPUT_LOGGER.log('INFO', 'Searching for Secrets')
            for signature in [sig for sig in signature_list if sig.category == 'secrets']:
                for scope in signature.scope:
                    search(
                        slack_con,
                        signature,
                        timeframe,
                        scope,
                        verbose)
        else:
            OUTPUT_LOGGER.log('INFO', 'Searching for PII')
            for signature in [sig for sig in signature_list if sig.category == 'pii']:
                for scope in signature.scope:
                    search(
                        slack_con,
                        signature,
                        timeframe,
                        scope,
                        verbose)

        OUTPUT_LOGGER.log('SUCCESS', f'Slack Watchman finished execution - Execution time:'
                                     f' {str(datetime.timedelta(seconds=time.time() - start_time))}')

    except TimeoutError as e:
        OUTPUT_LOGGER.log('ERROR', e)
        OUTPUT_LOGGER.log('DEBUG', traceback.format_exc())
    except Exception as e:
        OUTPUT_LOGGER.log('CRITICAL', e)
        OUTPUT_LOGGER.log('DEBUG', traceback.format_exc())
        sys.exit(1)


if __name__ == '__main__':
    main()
